// Copyright (C) 2013 S&P Capital IQ
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.spcapitaliq.platformexecuting

import java.io.{File, IOException}
import scala.annotation.tailrec

import scalaz.\/
import scalaz.syntax.id._
import scalaz.syntax.std.option._

/** How to acquire and release temporary `R`s.
  *
  * @define R `R`
  */
sealed abstract class TempStrategy[R] {
  /** Pivot between `acquire` and `release`. */
  type Fid

  /** Allocate and create a temporary $R. */
  def acquire(basename: String): (Fid, R)

  /** Release an `acquire`d $R. */
  def release(id: Fid): Unit
}

/** How to acquire and release temporary files.
  *
  * @define R `File`
  */
abstract class TempFileStrategy extends TempStrategy[File] {self =>
  /** Invoke `f` on every acquired file. */
  final def tapAcquiredFile(f: (Fid, File) => Unit): TempFileStrategy{type Fid = self.Fid} =
    new TempFileStrategy {
      type Fid = self.Fid
      def acquire(bn: String) =
        self.acquire(bn) <| f.tupled
      def release(id: Fid) = self.release(id)
    }
}

/** Some default `TempFileStrategy`s and tools for building them. */
object TempFileStrategy {
  private[this] val names = """(?s)(.*)(\..*)""".r

  private[platformexecuting] def splitBasename(f: String): (String, String) =
    (names findFirstMatchIn f
       cata (m => (m group 1, m group 2), (f, ".tmp")))

  /** Make a tempfile with a name more-or-less compatible with
    * `basename`, association-wise.
    */
  def jioCreateTempFile(basename: String): File = {
    val (p, s) = splitBasename(basename)
    File.createTempFile(Some(p) filter (_.length >= 3) getOrElse "zzz",
                        s)
  }

  /** A strategy that uses `java.io.File`'s built-in temporary file
    * allocation, requesting deletion on Java image exit rather than
    * explicit management by `#release`.
    */
  object javaIo extends TempFileStrategy {
    type Fid = Unit

    def acquire(basename: String) = {
      val fil = jioCreateTempFile(basename)
      fil.deleteOnExit()
      ((), fil)
    }

    def release(id: Fid) = ()
  }

  /** A strategy that uses `java.io.File`'s built-in temporary file
    * allocation, explicitly deleting it on `#release`.
    */
  object javaIoDeleteNow extends TempFileStrategy {
    type Fid = File

    def acquire(basename: String) = {
      val fil = jioCreateTempFile(basename)
      (fil, fil)
    }

    def release(id: Fid) = id.delete()
  }
}

/** How to acquire and release temporary directories.
  *
  * @define R directory
  */
abstract class TempDirStrategy extends TempStrategy[File]

/** Some default `TempDirStrategy`s and tools for building them. */
object TempDirStrategy {
  private[platformexecuting]
  def jioCreateTempDir(basename: String, tries: Int = 20): String \/ File = {
    @tailrec def rec(tries: Int): Option[File] =
      if (tries <= 0) None else {
        val tf = TempFileStrategy.jioCreateTempFile(basename)
        tf.delete()
        if (tf.mkdir()) Some(tf) else rec(tries - 1)
      }
    rec(tries) \/> ("Couldn't acquire a unique temporary directory in %d tries"
                      format tries)
  }

  /** A strategy that uses `java.io.File`'s built-in temporary file
    * allocation to choose a directory location, failing to delete it.
    */
  object javaIoNeverDelete extends TempDirStrategy {
    type Fid = Unit

    def acquire(basename: String) =
      ((), jioCreateTempDir(basename)
           fold (e => throw new IOException(e), identity))

    def release(id: Fid) = ()
  }
}
